
from channel import DataChannel

if __name__=='__main__':
    impala_db_config = {
        "driver": "impala",
        "config": {
            "host": "10.63.82.218",
            "username": "work",
            "password": "TwkdFNAdS1nIikzk",
            "database": "default",
            "port": 21050,
            "auth_mechanism": "LDAP",
        }
    }
    pg_db_config = {
        "driver": "postgres",
        "config": {
            "host": "10.63.82.191",
            "username": "dw_rw",
            "password": "Yxsj@123",
            "database": "test",
            "port": 5432
        }
    }
    csv_db_config = {
        "driver": "csv",
        "config": {
            "filename": "data",
            # "path": ""
        }
    }
    # with DataChannel(impala_db_config, pg_db_config) as channel:
    #     channel.table("tmp.t2")
    #     # channel.sink_db.execute("truncate table medical.t2")
    #     channel.insert("medical.t2", 2, is_create=True)
    
    # with DataChannel(pg_db_config, impala_db_config) as channel:
    #     channel.query("select * from medical.t2 limit 9")
    #     channel.insert("tmp.t3", 2, is_create=True)
    
    # with DataChannel(pg_db_config, csv_db_config) as channel: 
    #     channel.query("select * from medical.dim_date limit 123")
    #     channel.insert(is_create=True)
        
    with DataChannel(csv_db_config, pg_db_config) as channel: 
        channel.query()
        channel.insert("medical.t_csv", 2, is_create=True)
        
        


from base import Connect
from pyhive import hive

class HiveConnect(Connect):
    
    def __init__(self, config) -> None:
        """
        config = {
            
        }
        """
        
        self.connect = hive.connect(config)


from base import Connect, DTYPE
from impala import dbapi
from itertools import chain
import re

class ImpalaConnect(Connect):

    def __init__(self, norm_config) -> None:
        """
        norm_config = {
            "host": "127.0.0.1",
            "username": "admin",
            "password": "admin123",
            "database": "test",
            "port": 5432,
            ......
        }
        """
        config = norm_config
        config["user"] = norm_config.pop("username")

        self.connect = dbapi.connect(**config)
        self.cursor = self.connect.cursor()

        self.std_schema_data = []
        self.dataset_comment = ""
        self.placeholders = ""

        self.dtype_map = {
            # 必须都要对应
            "tinyint": ["tinyint"],
            "int": ["int"],
            "bigint": ["bigint"],
            "float": ["float"],
            "double": ["double"],
            "decimal": ["decimal"],
            "char": ["char"],
            "varchar": ["varchar"],
            "text": ["string"],
            "date": ["timestamp"],
            "time": ["timestamp"],
            "datetime": ["timestamp"],
            "timestamp": ["timestamp"],
        }

    def execute(self, sql):
        self.cursor.execute(sql)

    def query(self, sql):
        self.set_query_schema(sql)
        self.cursor.execute(sql)
        return self.cursor.fetchmany

    def table(self, table_name):
        self.set_table_schema(table_name)
        self.cursor.execute(f"select * from {table_name}")
        return self.cursor.fetchmany

    def insert(self, table_name, data):
        # TODO
        # print(data)

        # 一批数据会产生一个文件 OPTIMIZE
        # one_data = list(chain.from_iterable(data))  # 二维列表转一维列表
        # n_placeholders = ",".join([f"({self.placeholders})"] * len(data))
        # self.cursor.execute("insert into {} values {}".format(table_name, n_placeholders), one_data)

        # 一条数据会产生一个文件 OPTIMIZE
        self.cursor.executemany("insert into {} values ({})".format(table_name, self.placeholders), data)

    def set_query_schema(self, sql):
        """ return dataset schema: col_name, data_type, comment"""
        self.cursor.execute("select * from (" + sql + ") t limit 0")
        for row in self.cursor.description:
            col_name = row[0]
            data_type = self.convert_std_dtype(DTYPE(str(row[1]).lower(), row[4], row[5]))
            comment = row[2]
            self.std_schema_data.append([col_name, data_type, comment])
        self.placeholders = ",".join(["%s"] * len(self.std_schema_data))

    def set_table_schema(self, table_name):
        """ return dataset schema: col_name, data_type, comment """
        
        def split_str_type(str_type : str) -> DTYPE:
            result = re.search(r"(.*)\((.*),(.*)\)", str_type)
            if result:
                col_type = result.group(1)
                precision = result.group(2)
                scale = result.group(3)
                return DTYPE(col_type, precision, scale)
            
            result = re.search(r"(.*)\((.*)\)", str_type)
            if result:
                col_type = result.group(1)
                precision = result.group(2)
                return DTYPE(col_type, precision)
            
            return DTYPE(str_type)
        
        sql = f"describe {table_name}"
        self.cursor.execute(sql)
        for row in self.cursor.fetchall():
            col_name = row[0]
            data_type = self.convert_std_dtype(split_str_type(row[1]))
            comment = row[2]
            self.std_schema_data.append([col_name, data_type, comment])
            # self.dataset_comment = row[3]  #  describe无法获取表备注 TODO
        self.placeholders = ",".join(["%s"] * len(self.std_schema_data))

    def convert_std_dtype(self, dtype) -> DTYPE:
        # print(dtype)
        if dtype.get() == "varchar":
            return DTYPE("string")
        if dtype.get() == "numeric":
            return DTYPE("decimal", 38, 4)
        
        for key in self.dtype_map:
            types = self.dtype_map[key]
            for type in types:
                if type == dtype.name:
                    dtype.name = key
                    return dtype
        print(f"ERROR: self to std: {dtype} ==> varchar(128)")
    
    def convert_self_dtype(self, dtype) -> DTYPE:
        for key in self.dtype_map:
            if key == dtype.name:
                dtype.name = self.dtype_map[key][0]
                return dtype
        print(f"ERROR: std to self: {dtype} ==> varchar(128)")

    def create_table(self, table_name):
        cols_list = []
        for row in self.std_schema_data:
            col_name = row[0]
            data_type = self.convert_self_dtype(row[1]).get()
            comment = f"comment '{row[2]}'" if row[2] else ""
            cols_list.append("    {:<30} {:<10} {}".format(
                col_name, data_type, comment))

        sql = f"create table if not exists {table_name} (\n" + \
            ",\n".join(cols_list) + \
            "\n);"
        if self.dataset_comment:
            sql += f"\ncomment '{self.dataset_comment}';"
        print(sql)
        self.cursor.execute(sql)



from itertools import chain
from base import Connect, DTYPE
import MySQLdb

class MysqlConnect(Connect):
        
    def __init__(self, norm_config) -> None:
        """
        norm_config = {
            "host": "127.0.0.1",
            "username": "admin",
            "password": "admin123",
            "database": "test",
            "port": 3306,
            ......
        }
        """
        config = norm_config
        config["user"] = norm_config.pop("username")
        
        self.connect = MySQLdb.connect(**config)
        self.cursor = self.connect.cursor()
        
        self.std_schema_data = []
        self.dataset_comment = ""
        self.placeholders = ""
        
        self.dtype_map = {
            "tinyint": ["tinyint", "smallint"],
            "int": ["int", "mediumint"],
            "bigint": ["bigint"],
            "float": ["float"],
            "double": ["double"],
            "decimal": ["decimal"],
            "char": ["char"],
            "varchar": ["varchar"],
            "text": ["text"],
            "date": ["date"],
            "time": ["time"],
            "datetime": ["datetime"],
            "timestamp": ["timestamp"],
        }

    def execute(self, sql):
        self.cursor.execute(sql)
        
    def query(self, sql):
        self.set_query_schema(sql)
        self.cursor.execute(sql)
        return self.cursor.fetchmany

    def table(self, table_name):
        self.set_table_schema(table_name)
        self.cursor.execute(f"select * from {table_name}")
        return self.cursor.fetchmany
    
    def insert(self, table_name, data):
        one_data = list(chain.from_iterable(data))  # 二维列表转一维列表
        print(one_data)
        n_placeholders = ",".join([f"({self.placeholders})"] * len(data))
        print(n_placeholders)
        self.cursor.execute("insert into {} values {}".format(table_name, n_placeholders), one_data)
        # self.cursor.executemany("insert into {} values ({})".format(table_name, self.placeholders), data)
    
    def set_query_schema(self, sql):
        type_sql = "select oid, typname from pg_type where oid < 10000"
        self.cursor.execute(type_sql)
        type_map = {}
        for row in self.cursor.fetchall():
            type_map[row[0]] = row[1]
        
        self.cursor.execute("select * from (" + sql + ") t limit 0")
        for row in self.cursor.description:
            col_name = row[0]
            data_type = self.convert_std_dtype(DTYPE(type_map[row[1]], row[4], row[5]))
            comment = ""
            self.std_schema_data.append([col_name, data_type, comment])
        self.placeholders = ",".join(["%s"] * len(self.std_schema_data))
    
    def set_table_schema(self, table_name):
        # sql result: [col_name, col_type, precision, scale, column_comment, table_comment]
        sql = f"""
        select t1.column_name as col_name,
               t1.data_type as col_type, 
               coalesce(t1.character_maximum_length, t1.numeric_precision) as `precision`,  
               t1.numeric_scale as scale,
               t1.column_comment,
               t2.table_comment
	    from information_schema.columns as t1
	    join information_schema.tables t2 on t1.table_schema=t2.table_schema and t1.table_name=t2.table_name
	    where concat(t1.table_schema, '.', t1.table_name) = '{table_name}'
	    order by t1.ordinal_position
        """
        self.cursor.execute(sql)
        for row in self.cursor.fetchall():
            col_name = row[0]
            data_type = self.convert_std_dtype(DTYPE(row[1], row[2], row[3]))
            comment = row[4]
            self.std_schema_data.append([col_name, data_type, comment])
            self.dataset_comment = row[5]
        self.placeholders = ",".join(["%s"] * len(self.std_schema_data))
    
    def convert_std_dtype(self, dtype) -> DTYPE:
        print(dtype)
        if dtype.get() in ["varchar", "unknown"]:
            return DTYPE("string")
        if dtype.get() == "numeric":
            return DTYPE("decimal", 38, 4)
        
        for key in self.dtype_map:
            types = self.dtype_map[key]
            for type in types:
                if type == dtype.name:
                    dtype.name = key
                    return dtype
        print(f"ERROR: self to std: {dtype} ==> varchar(128)")
    
    def convert_self_dtype(self, dtype) -> DTYPE:
        for key in self.dtype_map:
            if key == dtype.name:
                dtype.name = self.dtype_map[key][0]
                return dtype
        print(f"ERROR: std to self: {dtype} ==> varchar(128)")
    
    def create_table(self, table_name):
        cols_list = []
        cols_comment_list = []
        for row in self.std_schema_data:
            col_name = row[0]
            # dtype = self.convert_self_dtype(row[1])
            data_type = self.convert_self_dtype(row[1]).get()
            comment = row[2]
            cols_list.append("    {:<30} {}".format(col_name, data_type))
            if comment:
                cols_comment_list.append(f"\ncomment on column {table_name}.{col_name} is '{comment}';")
        
        sql = f"create table if not exists {table_name} (\n" + \
            ",\n".join(cols_list) + \
            "\n);"
        if self.dataset_comment:
            sql += f"\ncomment on table {table_name} is '{self.dataset_comment}';"
        if cols_comment_list:
            sql += "".join(cols_comment_list)
        print(sql)
        self.cursor.execute(sql)


from datetime import datetime
from loguru import logger

class DataChannel:
    
    def __init__(self, source_config, sink_config):
        # driver_map = {
        #     # "mysql": MySQLdb.connect(**sink_config.config),
        #     "oracle": "cx_Oracle.connect()",
        #     # "mssql": pymssql.connect(**source_config.config),
        #     # "postgres": psycopg2.connect(**source_config["config"]),
        #     "postgres": PostgresConnect,
        #     # "hive": hive.connect(**source_config["config"]),
        #     "impala": ImpalaConnect,
        # }
        
        # 动态导入, 主要是避免安装所有数据库的依赖包, 用什么数据库安装什么数据库即可
        self.source_driver = source_config["driver"]
        self.sink_driver = sink_config["driver"]
        source_model = __import__(f"db_{self.source_driver}")
        sink_model = __import__(f"db_{self.sink_driver}")
        source_class = getattr(source_model, self.source_driver.capitalize() + "Connect")
        sink_class = getattr(sink_model, self.sink_driver.capitalize() + "Connect")

        self.source_db = source_class(source_config["config"])
        self.sink_db = sink_class(sink_config["config"])
        self.dataset = []
        self.start_time = datetime.now()
        # print(self.start_time)  # __init__ 比 __enter__ 优先执行
        
    def source_execute(self, sql):
        logger.info("source db execute: " + sql)
        self.source_db.execute(sql)

    def sink_execute(self, sql):
        logger.info("sink db execute: " + sql)
        self.sink_db.execute(sql)
    
    def query(self, sql=None):
        self.dataset = self.source_db.query(sql)
        self.sink_db.std_schema_data = self.source_db.std_schema_data
        self.sink_db.dataset_comment = self.source_db.dataset_comment
        self.sink_db.placeholders = self.source_db.placeholders
    
    def table(self, table_name):
        self.dataset = self.source_db.table(table_name)
        self.sink_db.std_schema_data = self.source_db.std_schema_data
        self.sink_db.dataset_comment = self.source_db.dataset_comment
        self.sink_db.placeholders = self.source_db.placeholders
    
    def insert(self, sink_table=None, batch_size=1000, is_create=False, header=True):
        
        if is_create:
            if self.sink_driver in ["csv", "excel"]:
                self.sink_db.create_table(header=header)
            else:
                self.sink_db.create_table(sink_table)

        i = 0
        start_time = datetime.now()
        logger.info("starting insert ......")
        while True:
            data = self.dataset(batch_size)
            if data:
                n = len(data)
                j = i + n

                if self.sink_driver in ["csv", "excel"]:
                    self.sink_db.insert(data)
                else:
                    self.sink_db.insert(sink_table, data)
                    
                elapsed_time = datetime.now() - start_time
                duration = 1 if elapsed_time.seconds == 0 else elapsed_time.seconds
                speed = int(j / duration)
                # progress = "{:.2f}".format(j / row_count * 100)
                logger.info(f"insert into {sink_table} {j} data succeed, speed {speed} records/s, elapsed time {elapsed_time}")
                i += n
            else:
                break
            
    def __enter__(self):
        logger.debug("init at " + str(datetime.now()))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self.source_db, "cursor") and self.source_db.cursor:
            self.source_db.cursor.close()
            logger.debug(str(self.source_db.cursor) + " cursor closed")
        if hasattr(self.sink_db, "cursor") and self.sink_db.cursor:
            self.sink_db.cursor.close()
            logger.debug(str(self.sink_db.cursor) + " cursor closed")
        if hasattr(self.source_db, "connect") and self.source_db.connect:
            self.source_db.connect.commit()
            self.source_db.connect.close()
            logger.debug(str(self.source_db.connect) + " connect closed")
        if hasattr(self.sink_db, "connect") and self.sink_db.connect:
            self.sink_db.connect.commit()
            self.sink_db.connect.close()
            logger.debug(str(self.sink_db.connect) + " connect closed")
        logger.info(f"Total time: {datetime.now() - self.start_time}s")


from abc import ABC, abstractmethod


class Connect(ABC):

    @abstractmethod
    def execute(self, sql):
        pass

    @abstractmethod
    def query(self, sql):
        pass

    @abstractmethod
    def table(self, table_name):
        pass

    @abstractmethod
    def insert(self, table_name):
        pass


class DTYPE:

    def __init__(self, name: str, precision: int = None, scale: int = None):
        self.name = name
        self.precision = precision
        self.scale = scale

    def get(self) -> str:
        if self.precision and self.scale:
            return f"{self.name}({self.precision},{self.scale})"

        if self.precision:
            return f"{self.name}({self.precision})"

        return self.name

    def __str__(self) -> str:
        return f"name={self.name}, precision={self.precision}, scale={self.scale}"



import csv
from pathlib import Path

from base import Connect, DTYPE

class CsvConnect(Connect):
    def __init__(self, norm_config) -> None:
        """
        norm_config = {
            "path": "/home/",
            "filename": "data",
            "separator": ",",
            "quoting": "",  # 
            ......
        }
        """
        config = norm_config
        
        # self.connect = None
        self.cursor = None
        
        self.std_schema_data = []
        self.dataset_comment = ""
        self.placeholders = ""
        
        path = config.get("path") if config.get("path") else "."
        filename = config["filename"]
        separator = config.get("separator") if config.get("separator") else ","
        quote = csv.QUOTE_MINIMAL
        
        self.file = Path(path).joinpath(filename + ".csv")
        self.data = None
        
        # register
        csv.register_dialect("dialect", delimiter=separator, quoting=quote)
        
    def execute(self, sql):
        raise Exception("Not Supported")
        
    def query(self, sql):
        self.set_query_schema()
        self.cursor = open(self.file, "r", encoding="utf-8", newline="")
        reader = csv.reader(self.cursor, "dialect")
        reader.__next__()  # 跳过header
        self.data = iter(row for row in reader)
        return self.fetchmany
    
    def set_query_schema(self):
        """ return dataset schema: col_name, data_type, comment"""
        with open(self.file, "r", encoding="utf-8", newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for col_name in reader.fieldnames:
                self.std_schema_data.append([col_name, DTYPE("text"), ""])
        self.placeholders = ",".join(["%s"] * len(self.std_schema_data))
    
    def fetchmany(self, size):
        batch_data = []
        for _ in range(size):
            try:
                batch_data.append(next(self.data))
            except StopIteration:
                break
        return batch_data

    def table(self, table_name):
        raise Exception("Not Supported")
    
    def insert(self, data):
        with open(self.file, "a", encoding="utf-8", newline="") as f:
            writer = csv.writer(f, "dialect")
            writer.writerows(data)
    
    def create_table(self, header=True):
        with open(self.file, "w", encoding="utf-8", newline="") as f:
            writer = csv.writer(f, "dialect")
            if header:
                header_data = [x[0] for x in self.std_schema_data]
                writer.writerow(header_data)
